/*
This program is distributed under the terms of the 'MIT license'. The text
of this licence follows...

Copyright (c) 2005 J.D.Medhurst (a.k.a. Tixy)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/**
@file

@brief Implementation of the IMA ADPCM audio coding algorithm
*/

#include "comdefine.h"
#include "IMA_ADPCM.h"
//#include <fstream>
//#include <iostream>

short  Pv_Val, Si_Val;
short st_Pv_Val, st_Pv2_Val;
short st_Si_Val, st_Si2_Val;

static const uint16_t IMA_ADPCMStepTable[89] =
	{
        7,     8,     9,    10,    11,    12,    13,
       14,    16,    17,    19,    21,    23,    25,
       28,    31,    34,    37,    41,    45,    50,
       55,    60,    66,    73,    80,    88,    97,
      107,   118,   130,   143,   157,   173,   190,
      209,   230,   253,   279,   307,   337,   371,
      408,   449,   494,   544,   598,   658,   724,
      796,   876,   963,  1060,  1166,  1282,  1411,
     1552,  1707,  1878,  2066,  2272,  2499,  2749,
     3024,  3327,  3660,  4026,  4428,  4871,  5358,
     5894,  6484,  7132,  7845,  8630,  9493, 10442,
    11487, 12635, 13899, 15289, 16818, 18500, 20350,
    22385, 24623, 27086, 29794, 32767
	};


static const int IMA_ADPCMIndexTable[256] =
	{
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8,
    -1, -1, -1, -1, 2, 4, 6, 8
	};



int IMA_ADPCM::Decode(unsigned adpcm)
	{
	int stepIndex = StepIndex;
	int step = IMA_ADPCMStepTable[stepIndex];

	stepIndex += IMA_ADPCMIndexTable[adpcm&7];
	if(stepIndex<0)
		stepIndex = 0;
	else if(stepIndex>88)
		stepIndex = 88;
	StepIndex = stepIndex;

	int diff = step>>3;
	if(adpcm&4)
		diff += step;
	if(adpcm&2)
		diff += step>>1;
	if(adpcm&1)
		diff += step>>2;


	int predicedValue = PredictedValue;
	if(adpcm&8)
		predicedValue -= diff;
	else
		predicedValue += diff;
	if(predicedValue<-0x8000)
		predicedValue = -0x8000;
	else if(predicedValue>0x7fff)
		predicedValue = 0x7fff;
	PredictedValue = predicedValue;

	return predicedValue;
	}

unsigned IMA_ADPCM::Decode(int16_t* dst, const uint8_t* src, int srcOffset, unsigned srcSize)
	{
   
    PredictedValue = Pv_Val;
    StepIndex=Si_Val;
    
	// use given bit offset
	src += srcOffset>>3;
	// calculate pointers to iterate output buffer
	int16_t* out = dst;
	int16_t* end = out+(srcSize>>2);
	
	
	while(out<end)
		{

		// get byte from src
		unsigned adpcm = *src;

		// pick which nibble holds a adpcm value...
		if(srcOffset&4)
			{
			adpcm >>= 4;  // use high nibble of byte
			++src;		  // move on a byte for next sample
			}

		*out++ = Decode(adpcm);  // decode value and store it

		// toggle which nibble in byte to write to next
		srcOffset ^= 4;
		}
	// return number of bytes written to dst
	return (unsigned)out-(unsigned)dst;
	}

/*int IMA_ADPCM::RChannel(unsigned adpcm)
	{
	int stepIndex = StepRight;
	int step = IMA_ADPCMStepTable[stepIndex];

	stepIndex += IMA_ADPCMIndexTable[adpcm&7];
	if(stepIndex<0)
		stepIndex = 0;
	else if(stepIndex>88)
		stepIndex = 88;
	StepRight = stepIndex;

	int diff = step>>3;
	if(adpcm&4)
		diff += step;
	if(adpcm&2)
		diff += step>>1;
	if(adpcm&1)
		diff += step>>2;

	int predicedValue2 = PredictedRight;
	if(adpcm&8)
		predicedValue2 -= diff;
	else
		predicedValue2 += diff;
	if(predicedValue2<-0x8000)
		predicedValue2 = -0x8000;
	else if(predicedValue2>0x7fff)
		predicedValue2 = 0x7fff;
	PredictedRight = predicedValue2;

	return predicedValue2;
	} */	

unsigned IMA_ADPCM::Stereo(int16_t* st_dst, const uint8_t* st_src, int st_srcOffset, 
unsigned st_srcSize)
	{
	  st_src += st_srcOffset>>3;
	  int16_t* st_out = st_dst;
	  int16_t* st_end = st_out+(st_srcSize>>2);
	 
     PredictedValue = st_Pv_Val;
     StepIndex = st_Si_Val;
     PredictedRight = st_Pv2_Val; //for r_channel not written out -
     StepRight = st_Si2_Val;      //for r_channel not written out -

	while(st_out<st_end)
		{
 
       for(int a = 8; a>0; a--){ //left channel ( this one we use )
  
	unsigned st_adpcm = *st_src;

		// pick which nibble holds a adpcm value...
		if(st_srcOffset&4)
			{
			st_adpcm >>= 4;  // use high nibble of byte
			++st_src;		// move on a byte for next sample
			}

		*st_out++ = Decode(st_adpcm);  // decode value and store it
		//write out single channel only - this converts wav data to
		//mono, 22050 sample rate

		// toggle which nibble in byte to write to next
		st_srcOffset ^= 4;
		}
       
       for(int b = 8; b>0; b--){ //right channel (data skipped over)
        
  
		unsigned st_adpcm2 = *st_src;

		// pick which nibble holds a adpcm value...
		if(st_srcOffset&4)
			{
			st_adpcm2 >>= 4;  // use high nibble of byte
			++st_src;		  // move on a byte for next sample
			}

		//*st_out++ = RChannel(st_adpcm2);// decode value and store it
		//not written out as the mixing of the channels wasn't correct
		//and produced poor quality stereo

		// toggle which nibble in byte to write to next
		 st_srcOffset ^= 4;
		 
		 }	 
        }
    
	// return number of bytes written to dst
	return (unsigned)st_out-(unsigned)st_dst;
	}	
